#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "hidapi.h"
#include <stdbool.h>

#define VENDOR_ID 0x07c0
#define PRODUCT_ID 0x111A
#define MAX_STR 255

//Achsen sind bei Joysticks immer auf Mittelstellung
#define JW_OFFSET 32768
#define JW56FR1_INPUTREPORT_SIZE_DEFAULT 9          //1 report ID (must be 0) , 8 data byte
#define JW56FR1_OUTPUTREPORT_SIZE_DEFAULT 13        //6 Byte ACC, 6 Byte ROT, 1 Byte buttons
#define JW56FR1_OUTPUTREPORT_SIZE_SPEED 62          //


//Umrechnung der RAW Daten des Sensors laut Datenblatt des 6-Achs Sensors (ST LSM6DSM)
// mg/LSB
#define SENSITIV_2G     0.061
#define SENSITIV_4G     0.122
#define SENSITIV_8G     0.244
#define SENSITIV_16G    0.448

// mdps/LSB
#define SENSITIV_125     4.375
#define SENSITIV_245     8.75
#define SENSITIV_500     17.50
#define SENSITIV_1000    35
#define SENSITIV_2000    70

typedef struct jw56fr1_param
{
    unsigned char aRange;
    unsigned char aFilter;
    unsigned char rRange;
    unsigned char rFilter;
    unsigned char mode;
} jw56fr1_param;


int setParameter(hid_device *dev, unsigned char aRange, unsigned char aFilter, unsigned char rRange, unsigned char rFilter);
int setMode(hid_device *dev, unsigned char mode);
int setStore(hid_device *dev);
void getInfo(hid_device *dev);



int setParameter(hid_device *dev, unsigned char aRange, unsigned char aFilter, unsigned char rRange, unsigned char rFilter)
{
    unsigned char buffer[JW56FR1_INPUTREPORT_SIZE_DEFAULT] = {0};

    buffer[0] = 0x00;		//ReportID
	buffer[1] = 0x00;		
	buffer[2] = aRange;
	buffer[3] = aFilter;
	buffer[4] = rRange;
	buffer[5] = rFilter;

    return hid_write(dev, buffer, JW56FR1_INPUTREPORT_SIZE_DEFAULT);
}

int setMode(hid_device *dev, unsigned char mode)
{
    unsigned char buffer[JW56FR1_INPUTREPORT_SIZE_DEFAULT] = {0};

    buffer[0] = 0x00;		//ReportID
	buffer[1] = 0x04;		
	buffer[2] = mode;       //0x00: default , 0x01: high speed

    return hid_write(dev, buffer, JW56FR1_INPUTREPORT_SIZE_DEFAULT);
}

int setStore(hid_device *dev)
{
    unsigned char buffer[JW56FR1_INPUTREPORT_SIZE_DEFAULT] = {0};

    buffer[0] = 0x00;		//ReportID
	buffer[1] = 0x01;		//store ettings into eeprom

    return hid_write(dev, buffer, JW56FR1_INPUTREPORT_SIZE_DEFAULT);
}

void getInfo(hid_device *dev)
{
    int res;
     wchar_t wstr[MAX_STR];

    printf("--- --- --- --- --- --- --- --- --- --- ---\n");
    printf("Print device informations\n\n");
    
    if ((res = hid_get_manufacturer_string(dev, wstr, MAX_STR)) >= 0)
        printf("Manufacturer String: %ls\n", wstr);

    if ((res = hid_get_product_string(dev, wstr, MAX_STR)) >= 0)
        printf("Product String: %ls\n", wstr);

    if ((res = hid_get_serial_number_string(dev, wstr, MAX_STR)) >= 0)
        printf("Serial Number String: (%d) %ls\n", wstr[0], wstr);

    printf("--- --- --- --- --- --- --- --- --- --- ---\n");
}


int main(int argc, char** argv)
{
    struct hid_device_info *devs = NULL;
    struct hid_device_info *cur_dev = NULL;

    char* path_0;   //Default interface (read/write)
    char* path_1;   //High speed ACC (read only)
    char* path_2;   //High speed Rotation (read only)
    hid_device *device_0;
    hid_device *device_1;
    hid_device *device_2;

    int res;
    int jw_found = 0;
    jw56fr1_param param;    //struct for jw56fr1 params and other stuff

    //Init default values for params
    param.mode = 0x00;      //0x00 for normal speed, 0x01 for high speed mode
    param.aRange = 0x00;
    param.aFilter = 0x00;
    param.rRange = 0x01;
    param.rFilter = 0x00;

    //Argumente auslesen
    if(argc >= 2)
    {
        //Alle Argumente auswerten und
        for(int i=0; i < argc; i++)
        {
            //Mode
            if(strcmp(argv[i], "-m") == 0)
                param.mode = strtol(argv[i+1], NULL, 16);

            //Acceleration range (values found in the datasheet)
            if(strcmp(argv[i], "-aR") == 0)
                param.aRange = strtol(argv[i+1], NULL, 16);

            //Acceleration filter (values found in the datasheet)
            if(strcmp(argv[i], "-aF") == 0)
                param.aFilter = strtol(argv[i+1], NULL, 16);

            //Rotation range (values found in the datasheet)
            if(strcmp(argv[i], "-rR") == 0)
                param.rRange = strtol(argv[i+1], NULL, 16);

            //Rotation filter (values found in the datasheet)
            if(strcmp(argv[i], "-rF") == 0)
                param.rFilter = strtol(argv[i+1], NULL, 16);
        }
    }

    //search for devices and set path for each interface
    hid_init();
    devs = hid_enumerate(VENDOR_ID, PRODUCT_ID);
    cur_dev = devs;
    
    while (cur_dev)
    {
        switch(cur_dev->interface_number)
        {
            case 0: //default interface for acc and range
                path_0 = cur_dev->path;
                jw_found++;
            break;
            case 1: //acceleration interface (for high speed mode)
                path_1 = cur_dev->path;     
            break;
            case 2: //rotation interface (for high speed mode)
                path_2 = cur_dev->path;
            break;
            default:
            break;
        }
        cur_dev = cur_dev->next;
    }

    //exit of no joywarrior56fr1 found
    if (jw_found == 0)
    {
        printf("JoyWarrior not found\n");
        hid_free_enumeration(devs);
        hid_exit();
        return -1;
    }

    //get default interface
    if ((device_0 = hid_open_path(path_0)) == NULL)
    {
        printf("Open device_0 failed\n");
        return -1;
    }

    //get high speed acceleration interface
    if ((device_1 = hid_open_path(path_1)) == NULL)
    {
        printf("Open device_1 failed\n");
        return -1;
    }

    //get high speed rotation interface
    if ((device_2 = hid_open_path(path_2)) == NULL)
    {
        printf("Open device_1 failed\n");
        return -1;
    }

    hid_free_enumeration(devs);

    printf("Set parameter\n");
    res = setParameter(device_0, param.aRange, param.aFilter, param.rRange, param.rFilter);   //take a look into the datasheet of JW56FR1 for the values
    if(res == -1)
    {
        printf("Error on setParameter()\n");
        return -1; 
    }
    
    if(param.mode == 0x01)
        printf("Device mode: high speed (onle ACC in this sample)\n");
    else
        printf("Device mode: normal speed\n");
    
    //Set the device mode
    res = setMode(device_0, param.mode);
    if(res == -1)
    {
        printf("Error on setMode()\n");
        return -1; 
    }


    //Wait for key press to go on
    printf("Press ANY key to continue. Press CTRL+C to exit\n");
    getchar();

//    hid_set_nonblocking(device_0, 1);

    int xAcc_raw = 0;
    int yAcc_raw = 0;
    int zAcc_raw = 0;

    int xRot_raw = 0;
    int yRot_raw = 0;
    int zRot_raw = 0;

    unsigned char buf[JW56FR1_OUTPUTREPORT_SIZE_SPEED] = {0};


    if(param.mode == 0x00)
    {
        printf("Start measurement...\n");
         while (1)
        {
            //12 Byte Data, ACC, Rot, Btn
            if(hid_read(device_0, buf, JW56FR1_OUTPUTREPORT_SIZE_DEFAULT) == -1) //Error -> Exit
            {
                printf("Error during read\n");
                break;
            }
        
            //RAW Data
            xAcc_raw = (buf[1] << 8) | buf[0];
            yAcc_raw = (buf[3] << 8) | buf[2];
            zAcc_raw = (buf[5] << 8) | buf[4];
            xRot_raw = (buf[7] << 8) | buf[6];
            yRot_raw = (buf[9] << 8) | buf[8];
            zRot_raw = (buf[11] << 8) | buf[10];

            printf("X %d Y %d Z %d    X %d Y %d Z %d    Btn %02X\n", xAcc_raw, yAcc_raw, zAcc_raw, xRot_raw, yRot_raw, zRot_raw, buf[12]);
        }
    }

    if(param.mode == 0x01)
    {
        printf("Start measurement (high speed)...\n");
        while (1)
        {
            //62 Byte data only ACC values
            if(hid_read(device_1, buf, JW56FR1_OUTPUTREPORT_SIZE_SPEED) == -1) //Error -> Exit
            {
                printf("Error during read\n");
                break;
            }

             for(int i=0; i<JW56FR1_OUTPUTREPORT_SIZE_SPEED; i++)
                 printf("%02X ", buf[i]);
            
            printf("\n");
        }
    }
    
    //dont't forget to close and exit
    hid_close(device_0);
    hid_close(device_1);
    hid_close(device_2);
    hid_exit();

    return 0;
}
